// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_DelayMeter_hpp
#define anna_core_util_DelayMeter_hpp

#include <string>

namespace anna {

/**
   Facilita la medicion de los tiempos empleados las distintas partes de nuestra aplicacion.

   Permite calcular tiempos acumulados como tiempos individuales. Por ejemplo:

   \code

   #include <anna/core/DelayMeter.hpp>

   void foo () {
      DelayMeter <_TimeUnit> meter;

      goo ();
      _TimeUnit gooTime = meter.getValue ();

      hoo ();
      _TimeUnit goohooTime = meter.setControlPoint ();

      joo ();
      _TimeUnit jooTime = meter.getValue ();
   }
   \endcode

   Dnde _TimeUnit podria ser anna::Second, anna::Millisecond, anna::Microsecond
*/
template <class _TimeUnit> class DelayMeter {
public:
  /**
     Constructor
     Inicializa la cuenta de temporizacion.
  */
  DelayMeter() { a_timestamp = _TimeUnit::getTime(); }

  /**
   * Constructor copia.
   * Copia la cuenta de utilizacion de la instancia recibida como parametro.
   * \param other Instancia de la copiar los parametros para calcular el tiempo transcurrido.
   */
  DelayMeter(const DelayMeter& other) : a_timestamp(other.a_timestamp), a_topReference(other.a_topReference) { ;}

  /**
   * Inicializa la cuenta de temporizacion. Este metodo es invocado automaticamente desde el contructor la clase
   * por lo que si vamos usar esta instancia para tomar un unica medida no es necesario invocarlo.
   * \warning Elimina el punto de referencia temporal que puediera haberse establecido con #setTopReference.
   * \return El nmero de milisegundos transcurridos desde la ltima vez que inicializamos la cuenta de temporizacion.
   */
  void setControlPoint() throw() {
    a_timestamp = _TimeUnit::getTime();
    clearTopReference();
  }

  /**
   * Inicializa la cuenta de temporizacion. Este metodo es invocado automaticamente desde el contructor la clase
   * por lo que si vamos usar esta instancia para tomar un unica medida no es necesario invocarlo.
   * \warning Elimina el punto de referencia temporal que puediera haberse establecido con #setTopReference.
   * \return El nmero de milisegundos transcurridos desde la ltima vez que inicializamos la cuenta de temporizacion.
   *
   * \param timestamp Valor de referencia a establecer.
   */
  void setControlPoint(const _TimeUnit& timestamp) throw() {
    a_timestamp = timestamp;
    clearTopReference();
  }

  /**
   * Se da la posiblidad de establecer un punto de referencia temporal de forma
   * que cuando se invoque a DelayMeter::getValue, el calculo de la diferencia de tiempo
   * no se har entre la marca de tiempo y el tiempo actual, sino la marca de
   * tiempo y sta marca de referencia.
   *
   * Esta funcionalidad ha sido requerida para medir el tiempo de ejecucin "real"
   * de tareas que se ejecutan dentro de un thread. Ya que puede pasar un tiempo
   * indeterminado desde que se termina la tarea MT (momento en el que se establecer
   * la marca de tiempo) y el ncleo y dems partes pueden tener conocimiento de que
   * esa tarea ha sido finalidad.
   */
  void setTopReference(const _TimeUnit& topReference) throw() { a_topReference = topReference; }

  /**
   * Elimina el punto de referencia temporal.
   */
  void clearTopReference() throw() { a_topReference = _TimeUnit(0); }

  /**
   * Inicializa el valor del punto de referencia.
   */
  void clear() throw() { a_timestamp = 0; }

  /**
   * Devuelve el nmero de milisegundos transcurridos desde la ltima vez que inicializamos la cuenta de temporizacin.
   * Si se ha establecido un punto de referencia mediante #setTopReference, devolver la diferencia entre el
   * el punto de control y la referencia, en otro caso, devolver la diferencia entre el punto de control y el
   * momento actual.
   * \return El nmero de milisegundos transcurridos desde la ltima vez que inicializamos la cuenta de temporizacion.
   * \warning Si detecta algun fallo devolvera 0.
   */
  _TimeUnit getValue() const throw() {
    a_now = (a_topReference == _TimeUnit(0)) ? _TimeUnit::getTime() : a_topReference;
    return (a_now > a_timestamp) ? (a_now - a_timestamp) : _TimeUnit(0);
  }

  /**
   * Devuelve el nmero de milisegundos transcurridos desde la ltima vez que inicializamos la cuenta de temporizacin.
   * Si se ha establecido un punto de referencia mediante #setTopReference, devolver la diferencia entre el
   * el punto de control y la referencia, en otro caso, devolver la diferencia entre el punto de control y el
   * momento actual.
   * \param now Valor temporal tomado como referencia.
   * \return El nmero de milisegundos transcurridos desde la ltima vez que inicializamos la cuenta de temporizacion.
   * \warning Si detecta algun fallo devolvera 0.
   */
  _TimeUnit getValue(const _TimeUnit& now) const throw() {
    return ((a_now = now) > a_timestamp) ? (a_now - a_timestamp) : _TimeUnit(0);
  }

  /**
   * Devuelve el tiempo que se us como referencia al calcular el retardo en #getValue
   * \return El tiempo que se us como referencia al calcular el retardo en #getValue
   */
  const _TimeUnit& getNow() const throw() { return a_now; }

  /**
   * Operador copia.
   * \param other Instancia de la que copiar.
   */
  DelayMeter& operator= (const DelayMeter& other) throw() { a_timestamp = other.a_timestamp; a_topReference = other.a_topReference; return *this; }

  /**
   * Compara el retardo acumulado por esta instancia con el valor recibido.
   * \param left Valor numrico con el comparar.
   * \return \em true si el retardo acumulado es mayor que el parmetro recibido o \em false en otro caso.
   */
  bool operator> (const _TimeUnit& left) const throw() { return getValue() > left; }

  /**
   * Compara el retardo acumulado por esta instancia con el valor recibido.
   * \param left Valor numrico con el comparar.
   * \return \em true si el retardo acumulado es mayor que el parmetro recibido o \em false en otro caso.
   */
  bool operator< (const _TimeUnit& left) const throw() { return getValue() < left; }

  /**
   * Devuelve la cadena que muestra el tiempo medido por esta instancia.
   * \return la cadena que muestra el tiempo medido por esta instancia.
   */
  std::string asString() const throw() { return getValue().asString(); }

  /**
   * Devuelve la cadena de depuracin de esta instancia.
   * \param whatis Texto con el nombre lgico de esta instancia.
   * \return la cadena de depuracin de esta instancia.
   */
  std::string asDebugString(const char* whatis) const throw() {
    std::string result(whatis);
    result += " { TopReference: ";
    result += a_topReference.asString();
    result += " | TimeStamp: ";
    result += a_timestamp.asString();
    result += " | Now: ";
    result += a_now.asString();
    return result += " }";
  }

private:
  _TimeUnit a_topReference;
  _TimeUnit a_timestamp;
  mutable _TimeUnit a_now;
};

}

#endif


