// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_http_Transport_hpp
#define anna_http_Transport_hpp

#include <anna/core/DataBlock.hpp>

#include <anna/comm/Transport.hpp>
#include <anna/comm/TransportFactory.hpp>
#include <anna/comm/TransportFactoryImpl.hpp>

#include <anna/http/internal/Tokenizer.hpp>
#include <anna/http/Message.hpp>

namespace anna {

namespace http {

class EncodedBlock;

namespace parser {
class Abstract;
}

/**
   Clase generica para definir la capa de transporte del protocolo HTTP.

   \see Transport.
*/
class Transport : public comm::Transport {
public:
  /**
   * Constructor.
   * Slo debera usarse en caso de tener que usar el mtodo #externalDecode.
   */
  Transport();

  /**
   * Destructor.
   */
  virtual ~Transport() {;}

  /**
     Devuelve la longitud del cuerpo del mensaje asociado.
     \return la longitud del cuerpo del mensaje asociado.
     \warning Exclusivamente de uso interno.
  */
  int getContentLength() const throw() { return a_contentLength; }

  /**
     Devuelve el mensaje de entrada asociado.
     \return La instancia del mensaje de entrada asociado.
     \warning Exclusivamente de uso interno.
  */
  http::Message* getInputMessage() throw(RuntimeException);

  /**
     Establece la longitud del cuerpo del mensaje asociado.
     \param contentLength Longitud del cuerpo del mensaje asociado.
     \warning Exclusivamente de uso interno.
  */
  void setContentLength(const int contentLength) throw(RuntimeException) { a_contentLength = contentLength; }

  /**
     Establece la posicion donde comienza el cuerpo del mensaje asociado.
     \param bodyOffset Posicion donde comienza el cuerpor del mensaje.
     \warning Exclusivamente de uso interno.
  */
  void setBodyOffset(const int bodyOffset) throw(RuntimeException) { a_bodyOffset = bodyOffset; }

  /**
     Metodo que inicializa el estado de esta capa de transporte. Se invoca
     directamente por el nucleo cuando sea necesario.
  */
  void clear() throw();

  /**
   * Permite obtener una instancia de anna::http::Message correspondiente al bfer recibido como parmetro.
   * \param buffer Espacio de memoria que contiene el mensaje HTTP a interpretar.
   * \param length Longitud del bfer.
   * \Warning Al ser invocado desde el exterior de la plataforma no hay ninguna seccin crtica que la
   * proteja, porque lo que si se invoca en modo MT desde varios thread habr que establecer las secciones
   * crticas necesarias para que slo un thread pueda acceder a este mtodo.
   */
  const http::Message* externalDecode(const char* buffer, const int length) throw(RuntimeException);

  /**
     Instancia un mensaje HTTP de entrada del tipo determinado por \em type.
     \param type Tipo de mensaje HTTP detectado en la entrada.
     \return La instancia del mensaje creado.
     \internal
  */
  http::Message* allocateInputMessage(const Message::Type::_v type) throw(RuntimeException);

  /**
   * Devuelve el bloque de datos codificados asociados a esta instancia.
   * \warning Exclusivamente uso interno.
   */
  EncodedBlock* getEncodedBlock() throw();

  /**
     Separa el token recibido como parametro en distintos componentes basicos, usando
     como separadores los caracteres de espacios, tabs, etc.

     \param token Token sobre el que deseamos aplicar la separacion.

     \return Una instancia del Tokenizer con el que poder recorrer todos los componentes.
  */
  const Tokenizer& split(const Token& token) throw(RuntimeException);

  /**
     Separa el token recibido como parametro en distintos componentes basicos.

     \param token Token sobre el que deseamos aplicar la separacion.
     \param separator Separador usado para detectar los tokens. No puede ser NULL:

     \return Una instancia del Tokenizer con el que poder recorrer todos los componentes.
  */
  const Tokenizer& split(const Token& token, const char* separator) throw(RuntimeException);

  /**
     Separa el token recibido como parametro en distintos componentes basicos.

     \param token Token sobre el que deseamos aplicar la separacion.
     \param separator Separador usado para detectar los tokens.

     \return Una instancia del Tokenizer con el que poder recorrer todos los componentes.
  */
  const Tokenizer& split(const Token& token, const char separator) throw(RuntimeException);

  /**
     Devuelve el literal que identifica de esta clase.
     \return el literal que identifica de esta clase.
  */
  static const char* className() throw() { return "anna::http::Transport"; }

  /**
     Devuelve el gestor de capas de transporte asociado a esta clase.
     \return El gestor de capas de transporte asociado a esta clase.
  */
  static comm::TransportFactory& getFactory() throw() { return st_factory; }

private:
  //-------------------------------------------------------------------------------------
  // - a_parser: Analizador del mensaje HTTP. Realiza un analisis en profundidad para
  // interpretar el contenido del mensaje.
  // - a_contentLenth: Tamao indicado por la etiqueta HTTP Content-length.
  // - a_bodyOffset: Indica el desplazamiento que hay que aplicar en http::Transport::decode
  // para direccionar el contenido del mensaje HTTP.
  // - a_result: El decode nos obliga a devolver un DataBlock => devolvemos el
  // a_inputMessage que nos dara la direccion al http::Message al invocar a
  // DataBlock::getData.
  // - a_haveToClear: Indica si el metodo calculeSize debe iniciar el estado de la instancia.
  // - a_inputMessage: Instancia del mensaje obtenida al realizar el analisis detallado
  // del ultimo mensaje recibido.
  //--------------------------------------------------------------------------------------
  const parser::Abstract* a_parser;
  int a_contentLength;
  int a_bodyOffset;
  DataBlock a_result;
  bool a_haveToClear;
  http::Message* a_inputMessage;
  Tokenizer a_fullScope;
  Tokenizer a_lineScope;
  EncodedBlock* a_encodedBlock;
  int a_lastChunkedByte;
  const DataBlock* a_fullMessage;

  static comm::TransportFactoryImpl <http::Transport> st_factory;

  void setParserState(const parser::Abstract* parser) throw(RuntimeException);

  int calculeSize(const DataBlock&) throw(RuntimeException);
  const comm::Message* decode(const DataBlock&) throw(RuntimeException);
  const DataBlock& code(comm::Message&) throw(RuntimeException);

  /*
     static const Token* tryRequest (const Tokenizer&) throw ();
     static const Token* tryResponse (const Tokenizer&) throw ();
  */
  friend class anna::Allocator <http::Transport>;
  friend class parser::Abstract;
};

}
}
#endif

