// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_ldap_Engine_hpp
#define anna_ldap_Engine_hpp

#include <map>
#include <string>

#include <anna/app/Component.hpp>

namespace anna {

namespace ldap {

class Session;

/**
   Gestor general de conexiones realizadas a diversos servidores LDAP.

   Optimiza la creacion, busqueda y liberacion de las sesines establecidas contra un
   numero indeterminado de servidores LDAP.

   El siguiente codigo muestra un ejemplo de implementacion:

   \code

      class MyEngine : public ldap::Engine {
      public:
         MyEngine () {;}

      private:
         anna::Recycler<MySession> a_sessions;

         anna::ldap::Session* allocateSession (const int category) throw () { return a_sessions.create (); }

         void releaseSession (anna::ldap::Session* session) throw () {
            MySession* aux = static_cast <MySession*> (session);
            a_sessions.release (aux);
         }
      };

   \endcode
*/
class Engine : public app::Component {
public:
  /**
   * Mscara de los niveles de depuracin que pueden ser usados en el mtodo #setDebugLevel
   * \see Engine
   */
  struct DebugLevel { enum _v { All = -1, None = 0 }; };

  /**
   * Devuelve el valor del indicador de conexin automtica. Por defecto este indicador ser \em true.
   * \return el valor del indicador de conexin automtica.
   */
  bool getAutoBind() const throw() { return a_autoBind; }

  /**
   * Establece el indicador de conexin automtica. En caso de no indicarse ser \em true.
   * \param autoBind Valor que tomar el indicador de conexin automtica.
   *
   * Si es necesario cambiar el temporizador del Bind de una sesin LDAP, primero habr que
   * crearla sin conexin automtica, cambiar el temporizador asociado e invocar al Bind invocando
   * implcitamente al mtodo ldap::Session::bind.
   */
  void setAutoBind(const bool autoBind) throw() { a_autoBind = autoBind; }

  /**
     Crea o reusa una sesin LDAP con los parmetros recibidos.

     Las sesines LDAP estaran definidas univocamente por la pareja (url, user) si al
     invocar a este metodo ya existiera una sesin identificada por los mismos parmetros
     recibidos, se devolvera su instancia.

     Si no existe una sesin identificada por la pareja (url, user) se creara mediate la llamada
     al metodo virtual puro (#allocateSession), se realizara la peticion de conexion y se devolvera
     esta nueva instancia.

     Dependiendo del indicador de conexin automtica se solicitar la conexin al servidor
     o ser el programador quien tenga que invocarlo mediate la llamada al mtodo ldap::Session::bind.

     \param url Direccion donde atiende peticiones el servidor LDAP.
     \param user Usuario requerido para establecer la conexion contra el servidor LDAP.
     \param password Password requerido para establecer la conexion contra el servidor LDAP.
     \param category Identifica el tipo de sesin a crear.

     \return La session identificada por la \em url y \em user indicados como parametro.

     \warning La conexion no estara totalmente operativa hasta que no se reciba la notificacion
     correspondiente en el metodo Session::eventResponse confirmando que el ClassCode::Bind se ha
     realizado correctamente.
  */
  Session* createSession(const char* url, const char* user, const char* password, const int category = 0)
  throw(RuntimeException);

  /**
     Crea o reusa una sesin LDAP con los parmetros recibidos.

     Las sesines LDAP estaran definidas univocamente por la pareja (url, id) si al
     invocar a este metodo ya existiera una sesin identificada por los mismos parmetros
     recibidos, se devolvera su instancia.

     Si no existe una sesin identificada por la pareja (url, id) se creara mediate la llamada
     al metodo virtual puro (#allocateSession), se realizara la peticion de conexion y se devolvera
     esta nueva instancia.

     Dependiendo del indicador de conexin automtica se solicitar la conexin al servidor
     o ser el programador quien tenga que invocarlo mediate la llamada al mtodo ldap::Session::bind.

     \param url Direccion donde atiende peticiones el servidor LDAP.
     \param id Identificador usado para identificar la sesin.
     \param user Usuario requerido para establecer la conexion contra el servidor LDAP.
     \param password Password requerido para establecer la conexion contra el servidor LDAP.
     \param category Identifica el tipo de sesin a crear.

     \return La session identificada por la \em url y \em id indicados como parametro.

     \warning La conexion no estara totalmente operativa hasta que no se reciba la notificacion
     correspondiente en el metodo Session::eventResponse confirmando que el ClassCode::Bind se ha
     realizado correctamente.
  */
  Session* createSession(const char* url, const int id, const char* user, const char* password, const int category = 0)
  throw(RuntimeException);

  /**
     Crea o reusa una sesin LDAP con los parmetros recibidos a un servidor que no requiera identificar
     el usuario que solicita la conexion.

     Las sesines LDAP estaran definidas univocamente por la pareja (url, user="") si al
     invocar a este metodo ya existiera una sesin identificada por los mismos parmetros
     recibidos, se devolvera su instancia.

     Si no existe una sesin identificada por la pareja (url, user="") se creara mediate la llamada
     al metodo virtual puro (#allocateSession), se realizara la peticion de conexion y se devolvera
     esta nueva instancia.

     \param url Direccion donde atiende peticiones el servidor LDAP.
     \param category Identifica el tipo de sesin a crear.

     \return La session identificada por la \em url indicada como parametro.

     \warning La conexion no estara totalmente operativa hasta que no se reciba la notificacion
     correspondiente en el metodo Session::eventResponse confirmando que el ClassCode::Bind se ha
     realizado correctamente.
  */
  Session* createSession(const char* url, const int category = 0) throw(RuntimeException) {
    return createSession(url, NULL, NULL, category);
  }

  /**
     Devuelve la instancia de la sesin identificada por la pareja (url, user) recibidos como parmetros.

     \param url Direccion de la maquina donde atiendo peticiones el servidor LDAP.
     \param user Usuario requerido para establecer la conexion contra el servidor LDAP.
     \param emode Modo de actuar en caso de que no exista una sesin que coincida con los parmetros indicados.

     \return La instancia de la sesin identificada por la pareja (url, user) recibidos como parmetros.

     \warning Si no hay ninguna sesin identificada por la pareja (url, user) se devolvera una excepcin.
  */
  Session* findSession(const char* url, const char* user, Exception::Mode::_v emode = Exception::Mode::Throw) throw(RuntimeException);

  /**
     Devuelve la instancia de la sesin identificada por la pareja (url, id) recibidos como parmetros.

     \param url Direccion de la maquina donde atiendo peticiones el servidor LDAP.
     \param id Identificador indicado para establecer la conexion contra el servidor LDAP.
     \param emode Modo de actuar en caso de que no exista una sesin que coincida con los parmetros indicados.

     \return La instancia de la sesin identificada por la pareja (url, id) recibidos como parmetros.

     \warning Si no hay ninguna sesin identificada por la pareja (url, id) se devolvera una excepcin.
  */
  Session* findSession(const char* url, const int id, Exception::Mode::_v emode = Exception::Mode::Throw) throw(RuntimeException);

  /**
     Devuelve la instancia de la sesin identificada por la pareja (url, user="") recibidos como parmetros.

     \param url Direccion de la maquina donde atiendo peticiones el servidor LDAP.
     \param emode Modo de actuar en caso de que no exista una sesin que coincida con los parmetros indicados.

     \return La instancia de la sesin identificada por la pareja (url, user="") recibidos como parmetros.

     \warning Si no hay ninguna sesin identificada por la pareja (url, user) se devolvera una excepcin.
  */
  Session* findSession(const char* url, Exception::Mode::_v emode = Exception::Mode::Throw) throw(RuntimeException) {
    return findSession(url, (const char*) NULL, emode);
  }

  /**
   * Libera todos los recursos asociados a una sesin LDAP, creada previemante con createSession. Si la sesin
   * fuera NULL esta operacin no tendr ningn efecto.
   * \param session Session LDAP a liberar.
   */
  void closeSession(Session* session) throw(RuntimeException);

  /**
     Devuelve un documento XML con la informacion relevante sobre esta instancia.
     \param parent Nodo XML del que colgar la informacion referente a esta instancia.
     \return Un documento XML con la informacion relevante sobre esta instancia.
  */
  virtual xml::Node* asXML(xml::Node* parent) const throw();

  /**
     Devuelve el nombre lgico de este anna::app::Component.
     \return El nombre lgico de este anna::app::Component.
  */
  static const char* getClassName() throw() { return "anna::ldap::Engine"; }

  /**
   * Establece el nivel de depuracin de las operaciones del OpenLDAP.
   * \param level Mscara que indica los elementos a depurar. Bsicamente 0 para desactivar las trazas de depuracin y -1 para
   * activar el trazado de todo.
   * \return El nivel de depuracin anterior.
   * \see http://www.openldap.org/doc/admin23/runningslapd.html para ms niveles.
   * \see DebugLevel
   */
  static int setDebugLevel(const int level) throw(RuntimeException);

protected:
  /**
     Constructor.
  */
  Engine();

  /**
     Metodo invocado para instanciar sesines.

     Para la creacion y liberacion de sesines es muy aconsejable usar el patron anna::Recycler.

     \param category Identifica la categoria o clase de sesin que deseamos instanciar. Este paremetro
     es el mismo indicado en el #createSession. Facilita que una misma aplicacion pueda crear un numero
     indeterminado de sesines de distintos tipos, es decir, que actuen de forma distinta a la hora
     de recoger los resultados de las peticiones.

     \see anna::Recycler
  */
  virtual Session* allocateSession(const int category) throw() = 0;

  /**
     Metodo invocado para liberar sesines. En caso de que nuestra aplicacion requiera varios tipos de
     sesines LDAP habra que tener en cuenta el valor devuelto por ldap::Session::getCategory y liberar
     el tipo adecuado de sesin.
     \see anna::Recycler
  */
  virtual void releaseSession(Session*) throw() = 0;

private:
  typedef std::pair <std::string, std::string> session_key;
  typedef std::map <session_key, Session*> session_container;
  typedef session_container::value_type session_value_type;
  typedef session_container::iterator session_iterator;
  typedef session_container::const_iterator const_session_iterator;

  session_container a_sessions;
  std::string a_auxURL; // Usada para contenar el ldap:// si fuera necesario
  bool a_autoBind;

  void do_initialize() throw() {;}
  void do_stop() throw();

  const char* completeURL(const char* url) throw();

  session_iterator session_find(const char* url, const char* user) throw() {
    return a_sessions.find(session_key(url, user));
  }
  session_iterator session_find(const std::string& url, const std::string& user) throw() {
    return a_sessions.find(session_key(url, user));
  }
  session_iterator session_find(const char* url, const int id) throw();

  session_iterator session_begin() throw() { return a_sessions.begin(); }
  session_iterator session_end() throw() { return a_sessions.end(); }
  static Session* session(session_iterator ii) throw() { return ii->second; }

  const_session_iterator session_begin() const throw() { return a_sessions.begin(); }
  const_session_iterator session_end() const throw() { return a_sessions.end(); }
  static const Session* session(const_session_iterator ii) throw() { return ii->second; }

  // Para el truco de poner a la alarma antes de invocar al ldap_result
  static void alarmnCatcher(int) throw();

};

}
}

#endif

