// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_xml_Node_hpp
#define anna_xml_Node_hpp

#include <vector>

#include <anna/core/Allocator.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/util/Recycler.hpp>
#include <anna/core/util/SortedVector.hpp>

#include <anna/xml/Namespace.hpp>

namespace anna {

namespace xml {

class Parser;
class Attribute;
class Text;
class XPath;
class Decompressor;

/**
   Nodo de documento XML.

   Cada nodo puede tener una serie indeterminada de atributos y nodos hijos. Por ejemplo:

   \code

   <xvc HeartBeat="20000">
      <broadcast>
         <INetAddress Address="204.152.65.15" Port="2000"/>
         <INetAddress Address="204.152.65.47" Port="2002"/>
      </broadcast>

      <ethernet Mode="raw" VirtualAddress="1.1.1.100">
         <Input Device="/dev/qfe" PhysicalAccessPoint="2" MAC="8:0:20:9e:ee:21"/>
         <Output Device="/dev/qfe" PhysicalAccessPoint="2" MAC="8:0:20:9e:ee:c9"/>
      </ethernet>
   </xvc>

   \endcode

   El nodo \em xvc tiene un atributo (ver Attribute) llamado \em HeartBeat y dos nodos
   hijos (\em broadcast y \em ethernet).
*/
class Node {
  struct NamespaceByName {
    static const std::string& value(const Namespace* ns) throw() { return ns->getName(); }
  };

public:
  typedef std::vector <Node*> Children; /**< Estructura usada para guardar los nodos hijos */
  typedef Children::iterator child_iterator; /**< Iterador sobre lista de nodos */
  typedef Children::const_iterator const_child_iterator; /**< Iterador sobre lista de nodos no modificables */

  typedef std::vector <Attribute*> Attributes; /**< Estructura usada para guardar los attributos */
  typedef Attributes::iterator attribute_iterator; /**< Iterador sobre la lista de atributos no modificables. */
  typedef Attributes::const_iterator const_attribute_iterator; /**< Iterador sobre la lista de atributos no modificables */

  typedef SortedVector <Namespace, NamespaceByName, std::string> namespace_container;
  typedef namespace_container::iterator namespace_iterator;
  typedef namespace_container::const_iterator const_namespace_iterator;

  /**
     Constructor.
     \param name Nombre del nodo.
  */
  Node(const char* name);

  /**
     Destructor.
  */
  virtual ~Node();

  /**
     Devuelve el nombre del nodo.
     \return El nombre del nodo.
  */
  const char* getName() const throw() { return a_name.c_str(); }

  /**
   * Devuelve el nombre del nodo.
   * \return El nombre del nodo.
   */
  const std::string& getNameAsString() const throw() { return a_name; }

  /**
     Devuelve la referencia al nodo predecesor de este nodo. Pueder ser NULL.
     \return La referencia al nodo predecesor de este nodo. Pueder ser NULL.
  */
  const Node* getParent() const throw() { return a_parent; }

  /**
     Devuelve la referencia al nodo predecesor de este nodo. Pueder ser NULL.
     \return La referencia al nodo predecesor de este nodo. Pueder ser NULL.
  */
  Node* getParent() throw() { return const_cast <Node*>(a_parent); }

  /**
     Devuelve la referencia al atributo que coincide con el nombre dado, si no existe lanzara
     una excepcion de ejecucion.
     \param name Nombre del atributo buscado.
     \param exceptionWhenNotFound Indica el comportamiento en caso de no encontrar el atributo que
     coincida con el nombre buscado.
     \return La referencia al atributo que coincide con el nombre dado.
  */
  const Attribute* getAttribute(const char* name, const bool exceptionWhenNotFound = true) const throw(anna::RuntimeException);

  /**
     Devuelve la referencia al texto asociado a este nodo. Pueder ser NULL.
     \return La referencia al texto asociado a este nodo. Pueder ser NULL.
  */
  const Text* getText() const throw() { return a_text; }

  /**
   * Devuelve el namespace asociado a este nodo. Puede ser NULL.
   * \return el namespace asociado a este nodo. Puede ser NULL.
   */
  const Namespace* getNamespace() const throw() { return a_namespace; }

  /**
     Devuelve un iterador al comienzo de la lista de nodo hijos.

     A continuacion presentamos un ejemplo de como se recorreria lista de nodos
     hijos de una determinado nodo.

     \code

     using namespace anna::xml;

     Node* theNode = ... <establece un valor correcto> ...
     Node* theChild;

     for (Node::iterator ii = theNode->begin (), maxii = theNode->end (); ii != maxii; ii ++) {
        theChild = Node::node (ii);

        .... <Procesar el hijo> ....
     }

     \endcode

     \return Un iterador al comienzo de la lista de nodo hijos.
  */
  child_iterator child_begin() throw() { return a_children.begin(); }

  /**
     Devuelve un iterador al comienzo de la lista de nodos hijos no modificables.

     A continuacion presentamos un ejemplo de como se recorreria lista de nodos
     hijos de una determinado nodo no modificable.

     \code

     using namespace anna::xml;

     const Node* theNode = ... <establece un valor correcto> ...
     const Node* theChild;

     for (Node::const_iterator ii = theNode->begin (), maxii = theNode->end (); ii != maxii; ii ++) {
        theChild = Node::node (ii);

        .... <Procesar el hijo> ....
     }

     \endcode

     \return Un iterador al comienzo de la lista de nodo hijos no modificables.
  */
  const_child_iterator child_begin() const throw() { return a_children.begin(); }

  /**
     Devuelve un iterador al comienzo de la lista de atributos no modificables.
     \return Un iterador al comienzo de la lista de atributos no modificables.
  */
  const_attribute_iterator attribute_begin() const throw() { return a_attributes.begin(); }

  /**
     Devuelve un iterador al comienzo de la lista de namespaces no modificables.
     \return Un iterador al comienzo de la lista de namespaces no modificables.
  */
  const_namespace_iterator namespace_begin() const throw() { return a_root->a_namespaces->begin(); }


  /**
     Devuelve un iterador al final de la lista de nodo hijos.
     \return Un iterador al final de la lista de nodo hijos.
     \see #child_begin
  */
  child_iterator child_end() throw() { return a_children.end(); }

  /**
     Devuelve un iterador al final de la lista de nodo hijos no modificables.
     \return Un iterador al final de la lista de nodo hijos no modificables.
     \see #begin
  */
  const_child_iterator child_end() const throw() { return a_children.end(); }

  /**
   * Devuelve el nmero de hijos definidos.
   * \return el nmero de hijos definidos.
   */
  int child_size() const throw() { return a_children.size(); }

  /**
     Devuelve un iterador al final de la lista de atributos no modificables.
     \return Un iterador al final de la lista de atributos no modificables.
  */
  const_attribute_iterator attribute_end() const throw() { return a_attributes.end(); }

  /**
     Devuelve un iterador al comienzo de la lista de namespaces no modificables.
     \return Un iterador al comienzo de la lista de namespaces no modificables.
  */
  const_namespace_iterator namespace_end() const throw() { return a_root->a_namespaces->end(); }

  /**
   * Devuelve el nmero de namespace definidos.
   * \return el nmero de namespace definidos.
   */
  int namespace_size() const throw() { return a_root->a_namespaces->size(); }

  /**
     Busca un nodo sucesor directo de este cuyo nombre coincida con el nombre recibido
     como parmetro.

     \param childName Nombre del nodo hijo que estamos buscando.
     \param exceptionWhenNotFound Indica el comportamiento en caso de no encontrar ningun nodo hijo que
     coincida con el nombre buscado.

     \return Referencia al nodo hijo que coincide con el nombre buscado. Si exceptionWhenNotFound es \em false
     y ninguno de los hijos cumple con la busqueda sera NULL.
  */
  const Node* find(const char* childName, const bool exceptionWhenNotFound = true) const
  throw(RuntimeException);

  /**
     Crea un atributo que depende de este nodo.
     \param name Nombre del atributo.
     \param value Valor asociado al atributo.
     \param _namespace Referencia al namespace al que pertenece el atributo. Puede ser NULL.
     \return La instancia del nuevo atributo.
  */
  xml::Attribute* createAttribute(const char* name, const char* value, const Namespace* _namespace = NULL) throw();

  /**
     Crea un atributo que depende de este nodo.
     \param name Nombre del atributo.
     \param value Valor asociado al atributo.
     \param _namespace Referencia al namespace al que pertenece el atributo. Puede ser NULL.
     \return La instancia del nuevo atributo.
  */
  xml::Attribute* createAttribute(const char* name, const std::string& value, const Namespace* _namespace = NULL) throw() {
    return createAttribute(name, value.c_str(), _namespace);
  }

  /**
     Crea un atributo que depende de este nodo.
     \param name Nombre del atributo.
     \param value Valor asociado al atributo.
     \param _namespace Referencia al namespace al que pertenece el atributo. Puede ser NULL.
     \return La instancia del nuevo atributo.
  */
  xml::Attribute* createAttribute(const char* name, const int value, const Namespace* _namespace = NULL) throw() {
    return createAttribute(name, anna::functions::asString(value), _namespace);
  }

  /**
     Crea un atributo que depende de este nodo.
     \param name Nombre del atributo.
     \param value Valor asociado al atributo. Pasamos como puntero para que no se creen ambiguedades
     con el codigo escrito hasta ahora.
     \param _namespace Referencia al namespace al que pertenece el atributo. Puede ser NULL.
     \return La instancia del nuevo atributo.
  */
  xml::Attribute* createAttribute(const char* name, const S64* value, const Namespace* _namespace = NULL) throw() {
    return createAttribute(name, anna::functions::asString(*value), _namespace);
  }

  /**
     Crea un atributo que depende de este nodo.
     \param name Nombre del atributo.
     \param value Valor asociado al atributo. Pasamos como puntero para que no se creen ambiguedades
     con el codigo escrito hasta ahora.
     \param _namespace Referencia al namespace al que pertenece el atributo. Puede ser NULL.
     \return La instancia del nuevo atributo.
  */
  xml::Attribute* createAttribute(const char* name, const U64* value, const Namespace* _namespace = NULL) throw() {
    return createAttribute(name, anna::functions::asString(*value), _namespace);
  }

  /**
     Crea el texto asociado a este nodo.
     \param text contain Valor del texto asociado a este nodo.
     \return La instancia del nuevo texto.
  */
  xml::Text* createText(const char* text) throw(RuntimeException);

  /**
     Crea el texto asociado a este nodo.
     \param text contain Valor del texto asociado a este nodo.
     \return La instancia del nuevo texto.
  */
  xml::Text* createText(const std::string& text) throw(RuntimeException) { return createText(text.c_str()); }

  /**
     Crea un nuevo nodo que depende de este.
     \param name Nombre del nodo hijo.
  */
  Node* createChild(const char* name) throw();

  /**
   * Crea un nuevo namespace (si procede) que podemos usar para asignar a los distintos nodos. Si el nodo ya existe y la referencia no coincide
   * con la registrada se obtendr una excepcin.
   *
   * \param name Nombre del nuevo namespace.
   * \param reference URI de la que obtener las definiciones.
   * \return La instancia de un namespace con los parmetros indicados.
   */
  const Namespace* createNamespace(const char* name, const char* reference) throw(RuntimeException) {
    std::string _name(name);
    return createNamespace(_name, reference);
  }

  /**
   * Crea un nuevo namespace (si procede) que podemos usar para asignar a los distintos nodos. Si el nodo ya existe y la referencia no coincide
   * con la registrada se obtendr una excepcin.
   *
   * \param name Nombre del nuevo namespace.
   * \param reference URI de la que obtener las definiciones.
   * \return La instancia de un namespace con los parmetros indicados.
   */
  const Namespace* createNamespace(const std::string& name, const char* reference) throw(RuntimeException);

  /**
   * Establece el namespace asociado a este nodo.
   * \param _namespace Instancia del namespace que vamos a asociar al nodo.
   */
  void setNamespace(const Namespace* _namespace) throw() { a_namespace = _namespace; }

  /**
   * Devuelve la instancia del namespace que coincide con el nombre recibido como parmetro. Puede ser NULL.
   * \return la instancia del namespace que coincide con el nombre recibido como parmetro. Puede ser NULL.
   */
  Namespace* namespace_find(const char* name,  const bool exceptionWhenNotFound = true) throw(RuntimeException) {
    const std::string _name(name);
    return namespace_find(_name, exceptionWhenNotFound);
  }

  /**
   * Devuelve la instancia del namespace que coincide con el nombre recibido como parmetro. Puede ser NULL.
   * \return la instancia del namespace que coincide con el nombre recibido como parmetro. Puede ser NULL.
   */
  Namespace* namespace_find(const std::string& name,  const bool exceptionWhenNotFound = true) throw(RuntimeException);

  /**
   * Devuelve la instancia del namespace que coincide con el nombre recibido como parmetro. Puede ser NULL.
   * \return la instancia del namespace que coincide con el nombre recibido como parmetro. Pueder ser NULL.
   */
  const Namespace* namespace_find(const char* name, const bool exceptionWhenNotFound = true) const throw(RuntimeException) {
    return const_cast <Node*>(this)->namespace_find(name, exceptionWhenNotFound);
  }

  /**
   * Devuelve la instancia del namespace que coincide con el nombre recibido como parmetro. Puede ser NULL.
   * \return la instancia del namespace que coincide con el nombre recibido como parmetro. Pueder ser NULL.
   */
  const Namespace* namespace_find(const std::string& name, const bool exceptionWhenNotFound = true) const throw(RuntimeException) {
    return const_cast <Node*>(this)->namespace_find(name, exceptionWhenNotFound);
  }

  /**
   * Libera todos los componentes (atributos, namespaces y nodos hijos) asociados a este nodo.
   */
  void clear() throw();

  /**
     Devuelve una cadena con toda la informacin relevante de esta instancia.
     \return Una cadena con toda la informacin relevante de esta instancia.
  */
  std::string asString() const throw();

  /**
     Devuelve la refencia al nodo sobre el que se encuentra el iterador pasado como
     parmetro.
     \param ii Iterador que estamos recorriendo.
     \return La refencia al nodo sobre el que se encuentra el iterador pasado como
     parmetro.
  */
  static Node* node(child_iterator& ii) throw() { return *ii; }

  /**
     Devuelve la refencia al nodo sobre el que se encuentra el iterador pasado como  parmetro.
     \param ii Iterador que estamos recorriendo.
     \return La refencia al nodo sobre el que se encuentra el iterador pasado como  parmetro.
  */
  static const Node* node(const_child_iterator& ii) throw() { return *ii; }

  /**
     Devuelve la refencia al atributo sobre el que se encuentra el iterador pasado como parmetro.
     \param ii Iterador que estamos recorriendo.
     \return La refencia al atributo sobre el que se encuentra el iterador pasado como parmetro.
  */
  static const Attribute* attribute(const_attribute_iterator& ii) throw() { return *ii; }

  /**
     Devuelve la refencia al namespace sobre el que se encuentra el iterador pasado como parmetro.
     \param ii Iterador que estamos recorriendo.
     \return La refencia al namespace sobre el que se encuentra el iterador pasado como parmetro.
  */
  static const Namespace* xnamespace(const_namespace_iterator& ii) throw() { return namespace_container::data(ii);; }

protected:
  /**
     Devuelve la refencia al atributo sobre el que se encuentra el iterador pasado como parmetro.
     \param ii Iterador que estamos recorriendo.
     \return La refencia al atributo sobre el que se encuentra el iterador pasado como parmetro.
  */
  static Attribute* attribute(attribute_iterator& ii) throw() { return *ii; }

private:
  std::string a_name;
  const Node* a_parent;
  Node* a_root;
  Children a_children;
  Attributes a_attributes;
  Text* a_text;
  const Namespace* a_namespace;

  // Esta instancia slo la crear el nodo root.
  namespace_container* a_namespaces;

  typedef Recycler <Node> node_pool;
  typedef Recycler <Attribute> attribute_pool;
  typedef Recycler <Text> text_pool;

  typedef Recycler <Namespace> namespace_pool;

  node_pool* a_node_pool;
  attribute_pool* a_attribute_pool;
  text_pool* a_text_pool;
  namespace_pool* a_namespace_pool;

  /* Para evitar que se pueda crear desde el exterior */
  Node();

  void setRoot(Node* root) throw() { a_root = root; }
  void setName(const char* name) throw() { a_name = name; }

  static const Attribute* find(const char* attrName, const_attribute_iterator, const_attribute_iterator) throw();

  // Allow sort in compiler
  attribute_iterator attribute_begin() throw() { return a_attributes.begin(); }
  attribute_iterator attribute_end() throw() { return a_attributes.end(); }

  friend class Parser;
  friend class Allocator<Node>;
  friend class XPath;
  friend class Decompressor;
  friend class Compiler;
};

}
}

#endif
