// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Local
#include <anna/statistics/Engine.hpp>
#include <anna/statistics/internal/sccs.hpp>
#include <anna/core/mt/Guard.hpp>
#include <anna/core/functions.hpp>
#include <anna/xml/xml.hpp>

// Std
#include <iostream>
#include <fstream>


using namespace anna::statistics;


//******************************************************************************
//----------------------------------------------------------------------- Engine
//******************************************************************************

//------------------------------------------------------------------------------
//------------------------------------------------------------- Engine::Engine()
//------------------------------------------------------------------------------
Engine::Engine() {
  statistics::sccs::activate();
  a_enabled = false;
  a_sequence_concept_id = 0;
}


//------------------------------------------------------------------------------
//------------------------------------------------------------ Engine::~Engine()
//------------------------------------------------------------------------------
Engine::~Engine() {
  for (_accumulator_map_it it = a_accumulators.begin(); it != a_accumulators.end(); it++)
    delete (it->second);
}


//------------------------------------------------------------------------------
//--------------------------------------------------------- Engine::addConcept()
//------------------------------------------------------------------------------
int Engine::addConcept(const std::string & description, const std::string & unit, const bool & integerNatureSample) throw() {
  a_sequence_concept_id++;
  _concept_identification_t aux;
  aux.SampleFile = ""; // sample dump disabled by default for new concepts
  aux.Description = description;
  aux.Unit = unit;
  aux.IntegerNatureSample = integerNatureSample;
  a_concept_identification_map[a_sequence_concept_id] = aux;
  return (a_sequence_concept_id);
}


//------------------------------------------------------------------------------
//--------------------------------------------------------- Engine::getConcept()
//------------------------------------------------------------------------------
bool Engine::getConcept(const int & id, std::string & description, std::string & unit, bool & integerNatureSample) const throw() {
  _concept_identification_map_iter it = a_concept_identification_map.find(id);

  if(it == a_concept_identification_map.end()) return false;

  description = (*it).second.Description;
  unit = (*it).second.Unit;
  integerNatureSample = (*it).second.IntegerNatureSample;
  return (true);
}


//------------------------------------------------------------------------------
//---------------------------------------------------- Engine::enableSampleLog()
//------------------------------------------------------------------------------
bool Engine::enableSampleLog(const int & id, const char *sampleFileName) throw() {
  _concept_identification_map_nc_iter it;
  std::string providedName = sampleFileName ? sampleFileName : "sample";
  std::string *SampleFile_ptr;
  std::string realName = "";

  if(id != -1) {
    it = a_concept_identification_map.find(id);

    if(it == a_concept_identification_map.end()) return false;

    if(providedName != "") realName = anna::functions::asString("%s.%d.csv", providedName.c_str(), id);

    // Assignment
    SampleFile_ptr = &((*it).second.SampleFile);
    *SampleFile_ptr = realName;
    return true;
  }

  // For all concepts:
  _concept_identification_map_nc_iter it_min(a_concept_identification_map.begin());
  _concept_identification_map_nc_iter it_max(a_concept_identification_map.end());
  for(it = it_min; it != it_max; it++) {
    realName = "";
    if(providedName != "") realName = anna::functions::asString("%s.%d.csv", providedName.c_str(), (*it).first);
    SampleFile_ptr = &((*it).second.SampleFile);
    *SampleFile_ptr = realName;
  }

  return true;
}


//------------------------------------------------------------------------------
//--------------------------------------------------- Engine::disableSampleLog()
//------------------------------------------------------------------------------
bool Engine::disableSampleLog(const int & id) throw() {

  if(id != -1) {
    _concept_identification_map_nc_iter it = a_concept_identification_map.find(id);

    if(it == a_concept_identification_map.end()) return false;

    // Access to map
    std::string *SampleFile_ptr = &((*it).second.SampleFile);
    *SampleFile_ptr = "";
    return true;
  }

  // For all concepts:
  return enableSampleLog(id /* -1 */, "");
}


//------------------------------------------------------------------------------
//--------------------------------------------------- Engine::disableSampleLog()
//------------------------------------------------------------------------------
bool Engine::logSample(const int & conceptId, const anna::Millisecond & unixTimestamp, const double & value) const throw() {
  anna::Guard guard(a_mutex);

  _concept_identification_map_iter it = a_concept_identification_map.find(conceptId);

  if(it == a_concept_identification_map.end()) return false;

  std::string target = (*it).second.SampleFile;

  if(target != "") {  // enabled
    std::ofstream out(target.c_str(), std::ifstream::out | std::ifstream::app);
    // Write and close (quiza no deberia cerrar cada vez...)
    std::string log = anna::functions::asString(unixTimestamp);
    log += ",";
    log += anna::functions::asString(value, (*it).second.IntegerNatureSample ? "%.0f" : "%e");
    log += "\n";
    out.write(log.c_str(), log.size());
    out.close();
  }

  return true;
}


//------------------------------------------------------------------------------
//-------------------------------------------------- Engine::createAccumulator()
//------------------------------------------------------------------------------
Accumulator *Engine::createAccumulator(const std::string &name) throw(anna::RuntimeException) {
  Accumulator *result = getAccumulator(name);

  if (result)
    throw anna::RuntimeException(anna::functions::asString("Cannot register another accumulator with the same name: %s", name.c_str()), ANNA_FILE_LOCATION);

  result = new Accumulator(name);
  a_accumulators[name] = result;
  return result;
}

Accumulator* Engine::getAccumulator(const std::string &name) throw() {
  Accumulator *result = NULL;

  _accumulator_map_nc_it it = a_accumulators.find(name);
  if (it != a_accumulators.end())
    result = it->second;

  return result;
}


//------------------------------------------------------------------------------
//----------------------------------------------------------- Engine::asString()
//------------------------------------------------------------------------------
std::string Engine::asString(void) const throw() {
  std::string trace;
  _concept_identification_map_iter iter;
  _concept_identification_map_iter iter_min(a_concept_identification_map.begin());
  _concept_identification_map_iter iter_max(a_concept_identification_map.end());
  trace =  "\n=============================";
  trace +=  "\nStatistic Engine Information";
  trace +=  "\n============================";

  if(!enabled())
    trace +=  "\nWARNING: engine is disabled !";

  if(a_concept_identification_map.size() == 0) {
    trace += "\nNo concepts found (empty map)";
    return trace;
  }

  trace += "\nNumber of elements (registered concepts) = "; trace += anna::functions::asString(a_concept_identification_map.size());

  for(iter = iter_min; iter != iter_max; iter++) {
    trace += "\n";
    trace += "\n   Concept id:            "; trace += anna::functions::asString((*iter).first);

    if((*iter).second.SampleFile != "") trace += "\n   Sample file:           "; trace += (*iter).second.SampleFile;

    trace += "\n   Description:           "; trace += (*iter).second.Description;
    trace += "\n   Unit:                  "; trace += (*iter).second.Unit;
    trace += "\n   Integer Nature Sample: "; trace += (*iter).second.IntegerNatureSample ? "yes" : "no";
  }

  trace += "\n";

  // Accumulators:
  trace += "\nNumber of accumulators= "; trace += anna::functions::asString(a_accumulators.size());
  for (_accumulator_map_it it = a_accumulators.begin(); it != a_accumulators.end(); it++) {
    trace += "\n";
    trace += it->second->asString();
  }

  return (trace);
}


//------------------------------------------------------------------------------
//-------------------------------------------------------------- Engine::asXML()
//------------------------------------------------------------------------------
anna::xml::Node* Engine::asXML(anna::xml::Node* parent, const int & numberOfDecimals) const throw() {
  anna::xml::Node* result = parent->createChild("anna.statistics.Engine");
  _concept_identification_map_iter iter;
  _concept_identification_map_iter iter_min(a_concept_identification_map.begin());
  _concept_identification_map_iter iter_max(a_concept_identification_map.end());
  int size = a_concept_identification_map.size();
  result->createAttribute("State", enabled() ? "Enabled" : "Disabled");
  anna::xml::Node* registeredConcepts = result->createChild("ConceptsList");
  registeredConcepts->createAttribute("Size", (size > 0) ? anna::functions::asString(size) : "No concepts found (empty map)");

  for(iter = iter_min; iter != iter_max; iter++) {
    anna::xml::Node* concept = registeredConcepts->createChild("Concept");
    concept->createAttribute("Id", anna::functions::asString((*iter).first));

    if((*iter).second.SampleFile != "") concept->createAttribute("SampleFile", (*iter).second.SampleFile);

    concept->createAttribute("Description", (*iter).second.Description);
    concept->createAttribute("Unit", (*iter).second.Unit);
    concept->createAttribute("IntegerNatureSample", (*iter).second.IntegerNatureSample ? "yes" : "no");
  }

  // Accumulators:
  anna::xml::Node* accumulators = result->createChild("anna.statistics.Accumulators");
  accumulators->createAttribute("RegisteredAccumulators", a_accumulators.size());
  for (_accumulator_map_it it = a_accumulators.begin(); it != a_accumulators.end(); it++)
    it->second->asXML(accumulators);

  return result;
}
