// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceWriter.hpp>
#include <anna/config/defines.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/mt/Guard.hpp>
#include <anna/xml/xml.hpp>

// Local
#include <anna/time/Date.hpp>
#include <anna/time/functions.hpp>
#include <anna/time/internal/sccs.hpp>

// Standard
#include <stdlib.h> // putenv
#include <iostream>


using namespace anna;
using namespace anna::time;



//******************************************************************************
//------------------------------------------------------------------------- Date
//******************************************************************************

// private:

//------------------------------------------------------------------------------
//----------------------------------------------------------- Date::initialize()
//------------------------------------------------------------------------------
void Date::initialize(const char * TZ) throw() {
  a_timestamp = ::time(NULL);
  setTz(TZ);
}


//------------------------------------------------------------------------------
//-------------------------------------------------------------- Date::refresh()
//------------------------------------------------------------------------------
void Date::refresh(void) throw() {
  anna::Guard guard(a_mutex);

  // Set current/programmed timezone
  a_tz.apply();

  // Rescue info:
  struct tm *ptrTm = localtime(&a_timestamp);
  a_tm = *ptrTm;
}



// public:

//------------------------------------------------------------------------------
//----------------------------------------------------------------- Date::Date()
//------------------------------------------------------------------------------
Date::Date(const char *TZ) {
  time::sccs::activate();

  if(!functions::initialized()) {
    std::cerr << std::endl << "Develop ERROR: you should firstly invoke anna::time::functions::initialize() before using time module" << std::endl;
  }

  initialize(TZ);
};


//------------------------------------------------------------------------------
//---------------------------------------------------------------- Date::setTz()
//------------------------------------------------------------------------------
void Date::setTz(const char * TZ) throw() {
  if (TZ) a_tz.set(TZ);
  else a_tz = functions::getSystemTimezone(); 

  refresh();
}


//------------------------------------------------------------------------------
//---------------------------------------------------- Date::setSystemTimezone()
//------------------------------------------------------------------------------
void Date::setSystemTimezone() throw() {
  a_tz = functions::getSystemTimezone();
  // Refresh the other data:
  refresh();
}


//------------------------------------------------------------------------------
//---------------------------------------------------------------- Date::store()
//------------------------------------------------------------------------------
void Date::store(const time_t & unixTimestamp) throw() {
  a_timestamp = unixTimestamp;
  // Refresh the other data:
  refresh();
}


//------------------------------------------------------------------------------
//------------------------------------------------------------- Date::storeNtp()
//------------------------------------------------------------------------------
void Date::storeNtp(const unsigned int & ntpTimestamp) throw() {
  a_timestamp = ntpTimestamp - TIMESTAMP_OFFSET_NTP1900_OVER_UNIX1970;
  // Refresh the other data:
  refresh();
}


//------------------------------------------------------------------------------
//---------------------------------------------------------------- Date::store()
//------------------------------------------------------------------------------
void Date::store(const struct tm &date, const char *TZ)

throw(anna::RuntimeException)

{
  if (TZ) {
    Timezone aux;
    aux.set(TZ);
    aux.apply();
  }
  else {
    functions::getSystemTimezone().apply();
  }

  // Set '-1' on 'tm_isdst', or it could be an ambiguity on tranformation
  //  returning inexact timestamp. This setting force system to check the correct flag:
  struct tm dateIsdst_1 = date;
  dateIsdst_1.tm_isdst = -1;
  time_t unixTimestamp = mktime((struct tm*) & dateIsdst_1);

  if(unixTimestamp == (time_t) - 1)
    throw anna::RuntimeException("Error during mktime() conversion !!", ANNA_FILE_LOCATION);

  a_timestamp = unixTimestamp;

  // Refresh the othr data:
  refresh();
}


//------------------------------------------------------------------------------
//---------------------------------------------------------------- Date::store()
//------------------------------------------------------------------------------
void Date::store(const std::string &stringDate, const char *TZ, const char *strptimeFormat)
throw(anna::RuntimeException)
{
  if(strptimeFormat == NULL)
    throw anna::RuntimeException("Invalid NULL strptimeFormat !!", ANNA_FILE_LOCATION);

  // Get equivalent 'tm':
  struct tm tmOrigin;
  memset(&tmOrigin, 0, sizeof(tmOrigin));

  if(strptime(stringDate.c_str(), strptimeFormat, &tmOrigin) == NULL) {
    std::string msg("Error during strptime() conversion: '");
    msg += stringDate;
    msg += "' can't be interpreted as '";
    msg += strptimeFormat;
    msg += "'";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  store(tmOrigin, TZ);
}


//------------------------------------------------------------------------------
//-------------------------------------------------------------- Date::operator=
//------------------------------------------------------------------------------
Date & Date::operator = (const Date & d) {
  // avoid itself copy: i.e., Date a,b; a=&b; a=b; b=a;
  if(this == &d) return (*this);

  a_timestamp = d.getUnixTimestamp();
  a_tz.set(d.getTz());
  a_tm = d.getTm();
  return (*this);
}


// gets


//------------------------------------------------------------------------------
//--------------------------------------------------------------- Date::getDay()
//------------------------------------------------------------------------------
const char *Date::getDay(void) const throw() {

  static const char *weekdayname[] = {"Sunday", "Monday", "Tuesday",
    "Wednesday", "Thursday", "Friday", "Saturday"};

  return weekdayname[a_tm.tm_wday];
}


//------------------------------------------------------------------------------
//------------------------------------------------------- Date::yyyymmddHHmmss()
//------------------------------------------------------------------------------
std::string Date::yyyymmddHHmmss(void) const throw() {
  return anna::functions::asString("%04d%02d%02d%02d%02d%02d",
     1900 + a_tm.tm_year, 1 + (a_tm.tm_mon), a_tm.tm_mday,
     a_tm.tm_hour, a_tm.tm_min, a_tm.tm_sec);
}


//------------------------------------------------------------------------------
//------------------------------------------------------------- Date::asString()
//------------------------------------------------------------------------------
std::string Date::asString(void) const throw() {

  std::string result = anna::functions::asString("%s %02d/%02d/%04d %02d:%02d:%02d ", getDay(),
                          getTm().tm_mday, 1 + (getTm().tm_mon), 1900 + getTm().tm_year,
                          getTm().tm_hour, getTm().tm_min, getTm().tm_sec);
  result += getTzAsString();
  result += anna::functions::asString(", isdst = %d [Unix Timestamp: %ld, Ntp Timestamp: %u]", getTm().tm_isdst, getUnixTimestamp(), getNtpTimestamp());
  result += ", System Timezone: ";
  result += functions::getSystemTimezone().asString();

  return result;
}


//------------------------------------------------------------------------------
//---------------------------------------------------------------- Date::asXML()
//------------------------------------------------------------------------------
anna::xml::Node* Date::asXML(anna::xml::Node* parent) const throw() {
  //anna::xml::Node* result = parent->createChild("anna.time.Date");
  parent->createAttribute("Date", anna::functions::asString("%02d/%02d/%04d", getTm().tm_mday, 1 + (getTm().tm_mon), 1900 + getTm().tm_year));
  parent->createAttribute("Day", getDay());
  parent->createAttribute("Time", anna::functions::asString("%02d:%02d:%02d", getTm().tm_hour, getTm().tm_min, getTm().tm_sec));
  parent->createAttribute("TZ", getTzAsString());
  parent->createAttribute("Isdst", (getTm().tm_isdst) ? "yes" : "no");
  parent->createAttribute("UnixTimestamp", anna::functions::asString((const int)getUnixTimestamp()));
  parent->createAttribute("NtpTimestamp", anna::functions::asString(getNtpTimestamp()));
  parent->createAttribute("SystemTimezone", functions::getSystemTimezone().asString());

  return parent;
}

